@@ -327,7 +327,7 @@ def vis_to_sph(vis_cube, sph_cube, umin, umax, nside, flagfile):
     out_cube.save(sph_cube)
 
 
-def do_gpr(i_cube, v_cube, gpr_config_i, gpr_config_v, rnd_seed=1):
+def do_gpr(i_cube, v_cube, gpr_config_i, gpr_config_v, rnd_seed=1, noise_cube=None):
     from ps_eor import fgfit
 
     t = time.time()
@@ -338,32 +338,40 @@ def do_gpr(i_cube, v_cube, gpr_config_i, gpr_config_v, rnd_seed=1):
 
     print('Running GPR for Stokes V ...\n')
 
+    gpr_v_noise = v_cube
+    if noise_cube is not None:
+        gpr_v_noise = noise_cube
+
     gpr_fit = fgfit.GprForegroundFit(gpr_config_v)
-    gpr_res_v = gpr_fit.run(v_cube, v_cube, rnd_seed=rnd_seed)
+    gpr_res_v = gpr_fit.run(v_cube, gpr_v_noise, rnd_seed=rnd_seed)
 
     print("\nDone in %.2f s\n" % (time.time() - t))
 
     t = time.time()
     print('Running GPR for Stokes I ...\n')
 
+    gpr_i_noise = gpr_res_v.sub
+    if noise_cube is not None:
+        gpr_i_noise = noise_cube
+
     gpr_fit = fgfit.ScaleNoiseGprForegroundFit(gpr_config_i)
-    gpr_res_i = gpr_fit.run(i_cube, gpr_res_v.sub, rnd_seed=rnd_seed)
+    gpr_res_i = gpr_fit.run(i_cube, gpr_i_noise, rnd_seed=rnd_seed)
 
     print("\nDone in %.2f s\n" % (time.time() - t))
 
     return gpr_res_i, gpr_res_v
 
 
-def do_ml_gpr(i_cube, noise_cube, ml_gpr_config, save_dir='.', save_name=None, rnd_seed=1):
+def do_ml_gpr(i_cube, noise_cube, ml_gpr_config, rnd_seed=1):
     from ps_eor import ml_gpr
 
-    # Saving PS stacker results for the residual, the noie
+    # Saving PS stacker results for the residual, the noise
     print('Running ML-GPR for Stokes I ...\n')
     t = time.time()
 
     fitter = ml_gpr.MLGPRForegroundFitter(ml_gpr_config)
 
-    ml_gpr_res = fitter.run(i_cube, noise_cube, save_dir=save_dir, save_name=save_name, verbose=True)
+    ml_gpr_res = fitter.run(i_cube, noise_cube, verbose=True)
 
     print(f"\nDone in {(time.time() - t):.2f} second\n")
 
@@ -437,8 +445,9 @@ def get_ratios(ps_gen, ps2d_ratio):
 @click.option('--plots_output_dir', '-po', help='Output directory for the figures')
 @click.option('--no_plot', help='Skip plotting', is_flag=True)
 @click.option('--rnd_seed', help='Set a random seed', default=3, type=int)
+@click.option('--noise_cube', 'file_noise_cube', help='Specify a noise cube (otherwise GPR residual of Stokes V is used)', default=None, type=t_file)
 def run_gpr(file_i, file_v, gpr_config_i, gpr_config_v, flag_config, eor_bins_list, ps_conf,
-            output_dir, plots_output_dir, no_plot, rnd_seed):
+            output_dir, plots_output_dir, no_plot, rnd_seed, file_noise_cube):
     ''' Run GPR & generate power spectra
 
     \b
@@ -477,6 +486,12 @@ def run_gpr(file_i, file_v, gpr_config_i, gpr_config_v, flag_config, eor_bins_li
             fig = flagger_runner.plot()
             fig.savefig(os.path.join(plots_output_dir, 'flagger_gpr.pdf'))
 
+    if file_noise_cube is not None:
+        noise_cube = datacube.DataCube.load(file_noise_cube)
+        noise_cube.filter_uvrange(ps_conf.umin, ps_conf.umax)
+        if flag_config:
+            noise_cube = flagger_runner.apply_last(noise_cube)
+
     ps_conf.el = 2 * np.pi * np.arange(i_cube.ru.min(), i_cube.ru.max(), ps_conf.du)
 
     for eor_name in ps_builder.eor_bin_list.get_all_names():
@@ -515,8 +530,12 @@ def run_gpr(file_i, file_v, gpr_config_i, gpr_config_v, flag_config, eor_bins_li
 
         print('Saving GPR result...\n')
 
+        eor_bin_noise_cube = None
+        if file_noise_cube is not None:
+            eor_bin_noise_cube = eor.get_slice_fg(noise_cube)
+
         gpr_res, gpr_res_v = do_gpr(eor.get_slice_fg(i_cube), eor.get_slice_fg(v_cube),
-                                    gpr_config_i, gpr_config_v, rnd_seed=rnd_seed)
+                                    gpr_config_i, gpr_config_v, noise_cube=eor_bin_noise_cube, rnd_seed=rnd_seed)
 
         gpr_res_v.save(out_dir, 'gpr_res_V')
         gpr_res.save(out_dir, 'gpr_res_I')
@@ -710,8 +729,10 @@ def run_ml_gpr(file_i, file_v, ml_gpr_config_file, flag_config, eor_bins_list, p
 
         i_cube_slice = eor.get_slice_fg(i_cube)
         n_cube_slice = eor.get_slice_fg(noise_cube)
-        ml_gpr_res = do_ml_gpr(i_cube_slice, n_cube_slice, ml_gpr_config, save_dir=out_dir,
-                               save_name='ml_gpr_res_I', rnd_seed=rnd_seed)
+        ml_gpr_res = do_ml_gpr(i_cube_slice, n_cube_slice, ml_gpr_config, rnd_seed=rnd_seed)
+
+        # Saving result
+        ml_gpr_res.save(out_dir, 'ml_gpr_res_I')
 
         # Get results
         scaled_noise_cube = ml_gpr_res.get_scaled_noise_cube()
@@ -747,12 +768,9 @@ def run_ml_gpr(file_i, file_v, ml_gpr_config_file, flag_config, eor_bins_list, p
         print('Plotting results ...\n')
 
         # Plotting MCMC results
-
-        ml_gpr_res.sampler.plot_corner(ml_gpr_config.mcmc.n_burn).savefig(
-            os.path.join(plot_out_dir, 'mcmc_corner.pdf'))
-        ml_gpr_res.sampler.plot_samples().savefig(os.path.join(plot_out_dir, 'mcmc_samples.pdf'))
-        ml_gpr_res.sampler.plot_samples_likelihood(n_burn=ml_gpr_config.mcmc.n_burn
-                                                   ).savefig(os.path.join(plot_out_dir, 'mcmc_samples_likelihood.pdf'))
+        ml_gpr_res.sampler_result.plot_corner().savefig(os.path.join(plot_out_dir, 'mcmc_corner.pdf'))
+        ml_gpr_res.sampler_result.plot_samples().savefig(os.path.join(plot_out_dir, 'mcmc_samples.pdf'))
+        ml_gpr_res.sampler_result.plot_samples_likelihood().savefig(os.path.join(plot_out_dir, 'mcmc_samples_likelihood.pdf'))
 
         # Plotting GPR results
         save_ps2d(ps_gen, ps_residual.get_ps(), ps_residual.get_ps2d(), 'Stokes I residual',
@@ -1032,14 +1050,25 @@ class FloatRangeType(click.ParamType):
         self.fail(f"{value!r} is not a valid float or a range of float", param, ctx)
 
 
+def validate_dec_deg(ctx, param, value):
+    if value is None:
+        return 90
+    if isfloat(value):
+        return float(value)
+    elif value == 'drift':
+        return value
+    click.BadParameter("dec_deg must be a float or 'drift'")
+
+
 @main.command('simu_uv')
-@click.argument('instru', type=click.Choice(['ska_low', 'nenufar', 'lofar_hba', 'nenufar_full', 'a12_hba'], case_sensitive=False))
+@click.argument('instru', type=click.Choice(['ska_low', 'nenufar', 'nenufar_80', 'nenufar_52', 'lofar_hba', 'a12_hba', 'a12_lba', 'mwa1'], case_sensitive=False))
 @click.argument('z', type=FloatRangeType())
 @click.option('--out_filename', '-o', help='Base filename of the output file', type=str, default='simu_uv',
               show_default=True)
 @click.option('--umin', help='Minimum baseline', default=None, type=int)
 @click.option('--umax', help='Maximum baseline', default=None, type=int)
-@click.option('--dec_deg', help='Target declination, in degrees', default=90, type=int, show_default=True)
+@click.option('--dec_deg', help='Target declination, in degrees (or "drift" for drift scan mode)', default=None, 
+              callback=validate_dec_deg, show_default=True)
 @click.option('--fov', help='Field of View side size (in deg), or "pb_fwhm" to set the FoV as the FWHM of the primary beam.',
               default=None, type=str)
 @click.option('--bandwidth', help='Bandwidth, in MHz', default=51 * 195.3 * 1e-3, type=float, show_default=True)
@@ -1073,8 +1102,16 @@ def simu_uv(instru, z, out_filename, umin, umax, dec_deg, fov, bandwidth, df, to
         freqs = np.arange(fmhz - bandwidth / 2, fmhz + bandwidth / 2, df * 1e-3) * 1e6
         z_str = f'{z:.1f}'
 
-    telescop = obssimu.TelescopeSimu.from_name(instru, freqs, dec_deg, 0, total_time, umin,
-                                               umax, int_time, not allow_intra_baselines)
+    telescop = obssimu.Telescope.from_name(instru)
+
+    if dec_deg == 'drift' or telescop.only_drift_mode:
+        dec_deg = telescop.latitude
+        if total_time > 0.25:
+            print('Warning: drift mode but total time > 15 min, are you sure ?')
+
+    # optimal observation crossing the meridian at half time
+    telescop_simu = obssimu.TelescopeSimu(telescop, freqs, dec_deg, - total_time / 2, total_time / 2, umin,
+                                          umax, int_time, not allow_intra_baselines)
 
     if fov is None:
         fov = telescop.fov
@@ -1090,20 +1127,21 @@ def simu_uv(instru, z, out_filename, umin, umax, dec_deg, fov, bandwidth, df, to
 
     print(f'Computing gridded UV configuration for instrument {telescop.name}')
     print(f'Simulating {total_time} hour long observation with {int_time} s integration time')
+    print(f'Pointing : {dec_deg:.1f} deg, HA between {- total_time / 2:.2f} hr and {total_time / 2:.2f} hr')
     print(f'Redshift z = {z_str}')
     print(f'{len(freqs)} frequency channels between {freqs.min() * 1e-6:.1f} and {freqs.max() * 1e-6:.1f} MHz')
     print(f'Frequency resolution: {df:.1f} kHz')
-    print(f'Baseline range: {telescop.umin} - {telescop.umax} lambda (discarding intra-baselines: {telescop.remove_intra_baselines})')
+    print(f'Baseline range: {telescop_simu.umin} - {telescop_simu.umax} lambda (discarding intra-baselines: {telescop_simu.remove_intra_baselines})')
     print(f'FoV: {fov:.1f} degrees')
 
-    uu_meter, vv_meter, _ = telescop.simu_uv()
+    uu_meter, vv_meter, _ = telescop_simu.simu_uv()
 
     if res is not None:
-        osamp_factor = int(np.round((1 / telescop.umax) / np.radians(6 / 60.)))
+        osamp_factor = int(np.round((1 / telescop_simu.umax) / np.radians(6 / 60.)))
     else:
         osamp_factor = 4
 
-    uv_simu = telescop.simu_gridded_uv(uu_meter, vv_meter, fov, osamp_factor, min_weight=0)
+    uv_simu = telescop_simu.simu_gridded_uv(uu_meter, vv_meter, fov, osamp_factor, min_weight=0)
 
     print(f'Image size: {uv_simu.weights.meta.shape}')
     print(f'Image resolution: {np.degrees(uv_simu.weights.meta.res) * 60:.2f} arcmin (oversampling: {osamp_factor})')
@@ -1160,9 +1198,10 @@ def simu_noise_img(uv_simu_h5, total_time_hour, out_filename, min_weight):
 @click.option('--filter_kpar_min', help='Filter mode below kpar_min', default=None, type=float)
 @click.option('--filter_wedge_theta', help='Filter mode below the wedge line', default=0, type=float)
 @click.option('--sefd', help='System Equivalent Flux Density (Jy)', default=None, type=float)
+@click.option('--n_incoherent_avg', help='Number of incoherent averaging', default=1, type=int)
 @click.option('--min_weight', help='Minimum number of visibilities in a uv-cell', default=10, type=int, show_default=True)
 def simu_noise_ps(uv_simu_h5, total_time_hour, out_filename, kmin, kmax, nks, 
-                  filter_kpar_min, filter_wedge_theta, sefd, min_weight):
+                  filter_kpar_min, filter_wedge_theta, sefd, n_incoherent_avg, min_weight):
     ''' Compute noise PS from a simulated UV coverage UV_SIMU_H5 for TOTAL_TIME_HOUR observation. '''
 
     from ps_eor import datacube, psutil, obssimu
@@ -1170,10 +1209,11 @@ def simu_noise_ps(uv_simu_h5, total_time_hour, out_filename, kmin, kmax, nks,
     uv_simu = obssimu.TelescopGriddedUV.load(uv_simu_h5)
     freqs = uv_simu.weights.freqs
     ps_gen = uv_simu.get_ps_gen(filter_kpar_min, filter_wedge_theta)
-    telescop = uv_simu.telescope_simu
+    telescop = uv_simu.telescope_simu.telescop
+    telescop_simu = uv_simu.telescope_simu
 
     print(f'Using telescope configuration {telescop.name}.')
-    print(f'Baseline range: {telescop.umin} - {telescop.umax} lambda (discarding intra-baselines: {telescop.remove_intra_baselines})')
+    print(f'Baseline range: {telescop_simu.umin} - {telescop_simu.umax} lambda (discarding intra-baselines: {telescop_simu.remove_intra_baselines})')
 
     if kmin is None:
         kmin = ps_gen.kmin
@@ -1195,6 +1235,11 @@ def simu_noise_ps(uv_simu_h5, total_time_hour, out_filename, kmin, kmax, nks,
     print('Generating power-spectra ...')
 
     for ps_name, ps in pss.items():
+        if n_incoherent_avg > 1:
+            ps.err = ps.err / np.sqrt(n_incoherent_avg)
+            ps.w = n_incoherent_avg * ps.w
+            ps.n_eff = n_incoherent_avg * ps.n_eff
+
         name = f'{out_filename}_{ps_name}_{uv_simu.name}_z{uv_simu.z:.1f}_{total_time_hour}h'
 
         print(f'Saving {ps_name} to {name} ...')
